using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Microsoft.Extensions.Logging;
using Orleans;
using Orleans.Configuration;
using Orleans.Hosting;
using System.IO;
using System.Threading.Tasks;

namespace OrleansDockerTest.Silo
{
    public class Program
    {
        public static async Task Main(string[] args)
        {
            var host = new HostBuilder()
                .ConfigureHostConfiguration(configHost =>
                {
                    configHost.SetBasePath(Directory.GetCurrentDirectory());
                    configHost.AddJsonFile("hostsettings.json", optional: true);
                    configHost.AddEnvironmentVariables();
                })
                .ConfigureAppConfiguration((hostContext, configApp) =>
                {
                    configApp.AddJsonFile("appsettings.json", optional: true);
                    configApp.AddJsonFile(
                        $"appsettings.{hostContext.HostingEnvironment.EnvironmentName}.json",
                        optional: true);
                    configApp.AddEnvironmentVariables();
                })
                .ConfigureServices((hostContext, services) =>
                {
                    var siloBuilder = BuildSilo(hostContext);
                    services.AddSingleton(typeof(ISiloHost), siloBuilder.Build());
                    services.AddHostedService<OrleansHostedService>();
                })
                .ConfigureLogging((hostContext, configLogging) =>
                {
                    configLogging.AddConsole();
                })
                .UseConsoleLifetime()
                .Build();

            await host.RunAsync();
        }

        private static ISiloHostBuilder BuildSilo(Microsoft.Extensions.Hosting.HostBuilderContext hostContext)
        {
            var connectionString = "Host=postgredb; Port=5432; Username=postgres; Password=1234qqqQ; Database=Apostil; Pooling=true;";
            var adoProvider = "Npgsql";

            var builder = new SiloHostBuilder()
                .ConfigureServices((siloContext, services) =>
                {
                })
                .ConfigureApplicationParts(parts =>
                {
                    parts.AddApplicationPart(typeof(TestGrain).Assembly).WithReferences();
                })
                .Configure<ClusterOptions>(options =>
                {
                    options.ClusterId = "Cluster";
                    options.ServiceId = "OrleansDockerTest";
                })
                .ConfigureLogging(logging =>
                {
                    logging.AddConsole();
                })
                .AddAdoNetGrainStorage("PubSubStore", o =>
                {
                    o.UseJsonFormat = true;
                    o.IndentJson = false;
                    o.ConnectionString = connectionString;
                    o.Invariant = adoProvider;
                })
                .AddAdoNetGrainStorageAsDefault(o =>
                {
                    o.UseJsonFormat = true;
                    o.IndentJson = false;
                    o.ConnectionString = connectionString;
                    o.Invariant = adoProvider;
                })
                .AddSimpleMessageStreamProvider("StreamProvider", o =>
                {
                    o.PubSubType = Orleans.Streams.StreamPubSubType.ImplicitOnly;
                    o.OptimizeForImmutableData = true;
                    o.FireAndForgetDelivery = false;
                })
                .ConfigureEndpoints(
                    siloPort: 11111,
                    gatewayPort: 30000)
                .UseAdoNetClustering(o =>
                {
                    o.ConnectionString = connectionString;
                    o.Invariant = adoProvider;
                })
                .UseAdoNetReminderService(o =>
                {
                    o.ConnectionString = connectionString;
                    o.Invariant = adoProvider;
                });

            return builder;
        }
    }
}
